import os.path as op
from typing import List

import geopandas as gpd
import numpy as np
import pandas as pd
from hydromt_sfincs import SfincsModel, utils

from bluemath_tk.wrappers.sfincs.sfincs_wrapper import SfincsModelWrapper


class SfincsPabloModelWrapper(SfincsModelWrapper):
    """
    Wrapper for the SFINCS model (Pablo version 10/02/2025).
    """

    p_data = "/home/tausiaj/GitHub-GeoOcean/BlueMath/test_data/SFINCS/data"

    def build_case(
        self,
        case_context: dict,
        case_dir: str,
        catalogs_list: List[str],
    ) -> None:
        """
        Build the input files for a case.

        Parameters
        ----------
        case_context : dict
            The case context.
        case_dir : str
            The case directory.
        catalogs_list : List[str]
            The list of catalogs.
        """

        # Instantiate the model
        sf_model = SfincsModel(data_libs=catalogs_list, root=case_dir, mode="w+")
        sf_model.setup_grid(
            x0=408002.5,
            y0=8467002.5,
            dx=5.0,
            dy=5.0,
            nmax=1659,
            mmax=3299,
            rotation=0,
            epsg=32702,
        )
        datasets_dep = [{"elevtn": op.join(self.p_data, "apia_dem_v2.tif")}]
        sf_model.setup_dep(datasets_dep=datasets_dep)
        sf_model.setup_mask_active(zmin=-4, reset_mask=True)
        sf_model.setup_mask_bounds(btype="waterlevel", zmax=-1.99, reset_bounds=True)
        sf_model.setup_mask_bounds(btype="outflow", zmin=1, reset_bounds=False)
        sf_model.setup_config(
            **{
                "tref": "20000101 000000",
                "tstart": "20000101 000000",
                "tstop": "20000101 022315",
                "dtout": "100",
            }
        )
        #### Load location of focings csv
        location_points = pd.read_csv(op.join(self.p_data, "forcing_points.csv"))
        # start_points = pd.read_csv(op.join(self.p_data, "start_points.csv"))
        # end_points = pd.read_csv(op.join(self.p_data, "end_points.csv"))
        #### Load forcing levels csv
        water_levels = pd.read_csv(
            op.join(
                self.p_data,
                f"water_levels_sfincs_apia_rp_{case_context.get('waterlevel')}.csv",
            ),
            index_col=0,
        )
        water_levels.columns = range(1, len(water_levels.columns) + 1)
        #### Loc points
        x = location_points.x.values
        y = location_points.y.values

        # add to Geopandas dataframe as needed by HydroMT
        pnts = gpd.points_from_xy(x, y)
        index = np.arange(1, len(x) + 1, 1)  # NOTE that the index should start at one
        bnd = gpd.GeoDataFrame(index=index, geometry=pnts, crs=sf_model.crs)
        # add time series
        time = pd.date_range(
            start=utils.parse_datetime(sf_model.config["tstart"]),
            end=utils.parse_datetime(sf_model.config["tstop"]),
            periods=len(water_levels),
        )
        water_levels.index = time
        # water_levels[2] = water_levels[1]
        sf_model.setup_waterlevel_forcing(timeseries=water_levels, locations=bnd[0::49])
        sf_model.write()  # write all

    def build_cases(
        self,
        catalogs_list: List[str],
        mode: str = "one_by_one",
    ) -> None:
        """
        Build the input files for all cases.

        Parameters
        ----------
        catalogs_list : List[str]
            The list of catalogs.
        mode : str, optional
            The mode to build the cases. Default is "one_by_one".

        Raises
        ------
        ValueError
            If the cases were not properly built
        """

        super().build_cases(mode=mode)
        if not self.cases_context or not self.cases_dirs:
            raise ValueError("Cases were not properly built.")
        for case_context, case_dir in zip(self.cases_context, self.cases_dirs):
            self.build_case(
                case_context=case_context,
                case_dir=case_dir,
                catalogs_list=catalogs_list,
            )


# Usage example
if __name__ == "__main__":
    # Define the input parameters
    templates_dir = (
        "/home/tausiaj/GitHub-GeoOcean/BlueMath/bluemath_tk/wrappers/sfincs/templates/"
    )
    output_dir = "/home/tausiaj/GitHub-GeoOcean/BlueMath/test_cases/sfincs/pablo/"
    # Load swan model parameters
    model_parameters = {"waterlevel": [5.0, 20.0, 50.0, 100.0]}
    # Create an instance of the SWAN model wrapper
    sfincs_wrapper = SfincsPabloModelWrapper(
        templates_dir=templates_dir,
        model_parameters=model_parameters,
        output_dir=output_dir,
    )
    # Build the input files
    sfincs_wrapper.build_cases(
        mode="one_by_one",
        catalogs_list=[
            "/home/tausiaj/GitHub-GeoOcean/BlueMath/test_data/SFINCS/catalogues/manning_cat.yml",
            "/home/tausiaj/GitHub-GeoOcean/BlueMath/test_data/SFINCS/catalogues/topo_cat.yml",
        ],
    )
    # List available launchers
    print(sfincs_wrapper.list_available_launchers())
    # Run the model
    sfincs_wrapper.run_cases(launcher="docker", parallel=True)



import copy
import itertools
import os
import subprocess
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import List, Union

import numpy as np
import xarray as xr
from jinja2 import Environment, FileSystemLoader

from ..core.models import BlueMathModel
from ._utils_wrappers import copy_files, write_array_in_file


class BaseModelWrapper(BlueMathModel):
    """
    Base class for numerical models wrappers.

    Attributes
    ----------
    templates_dir : str
        The directory where the templates are stored.
    model_parameters : dict
        The parameters to be used in the templates.
    output_dir : str
        The directory where the output files will be saved.
    env : Environment
        The Jinja2 environment.
    templates_name : List[str]
        The names of the templates.
    cases_dirs : List[str]
        The list with cases directories.
    cases_context : List[dict]
        The list with cases context.

    Methods
    -------
    _check_parameters_type -> None
        Check if the parameters have the correct type.
    _exec_bash_commands -> None
        Execute bash commands.
    list_available_launchers -> dict
        List the available launchers.
    set_cases_dirs_from_output_dir -> None
        Set the cases directories from the output directory.
    write_array_in_file -> None
        Write an array in a file.
    copy_files -> None
        Copy file(s) from source to destination.
    render_file_from_template -> None
        Render a file from a template.
    create_cases_context_one_by_one -> List[dict]
        Create an array of dictionaries with the combinations of values from the
        input dictionary, one by one.
    create_cases_context_all_combinations -> List[dict]
        Create an array of dictionaries with each possible combination of values
        from the input dictionary.
    build_cases -> None
        Create the cases folders and render the input files.
    run_case -> None
        Run the case based on the launcher specified.
    run_cases -> None
        Run the cases based on the launcher specified.
        Parallel execution is optional.
        Cases to run can be specified.
    run_cases_bulk -> None
        Run the cases based on the launcher specified.
    postprocess_case -> None
        Postprocess the model output for a specific case.
    join_postprocessed_files -> xr.Dataset
        Join the postprocessed files.
    postprocess_cases -> Union[xr.Dataset, List[xr.Dataset]]
        Postprocess the model output.
    """

    def __init__(
        self,
        templates_dir: str,
        model_parameters: dict,
        output_dir: str,
        templates_name: List[str] = "all",
        default_parameters: dict = None,
    ) -> None:
        """
        Initialize the BaseModelWrapper.
        """

        super().__init__()
        if default_parameters is not None:
            self._check_parameters_type(
                default_parameters=default_parameters, model_parameters=model_parameters
            )
        self.templates_dir = templates_dir
        self.model_parameters = model_parameters
        self.output_dir = output_dir
        self._env = Environment(loader=FileSystemLoader(self.templates_dir))
        if templates_name == "all":
            self.logger.warning(
                f"Templates name is 'all', so all templates in {self.templates_dir} will be used."
            )
            self.templates_name = self.env.list_templates()
            self.logger.info(f"Templates names: {self.templates_name}")
        else:
            self.templates_name = templates_name
        self.cases_dirs: List[str] = []
        self.cases_context: List[dict] = []

    @property
    def env(self) -> Environment:
        return self._env

    def _check_parameters_type(
        self, default_parameters: dict, model_parameters: dict
    ) -> None:
        """
        Check if the parameters have the correct type.
        This function is called in the __init__ method of the BaseModelWrapper,
        but default_parameters are defined in the child classes.
        This way, child classes can define default types for parameters.

        Parameters
        ----------
        default_parameters : dict
            The default parameters type for the model.
        model_parameters : dict
            The parameters to be used in the templates.

        Raises
        ------
        ValueError
            If a parameter has the wrong type.
        """

        for model_param, param_value in model_parameters.items():
            if model_param not in default_parameters:
                self.logger.warning(
                    f"Parameter {model_param} is not in the default_parameters"
                )
            else:
                if isinstance(param_value, (list, np.ndarray)) and all(
                    isinstance(item, default_parameters[model_param])
                    for item in param_value
                ):
                    self.logger.info(
                        f"Parameter {model_param} has the correct type: {default_parameters[model_param]}"
                    )
                else:
                    raise ValueError(
                        f"Parameter {model_param} has the wrong type: {default_parameters[model_param]}"
                    )

    @staticmethod
    def _exec_bash_commands(
        str_cmd: str, out_file: str = None, err_file: str = None
    ) -> None:
        """
        Execute bash commands.

        Parameters
        ----------
        str_cmd : str
            The bash command.
        out_file : str, optional
            The name of the output file. If None, the output will be printed in the terminal.
            Default is None.
        err_file : str, optional
            The name of the error file. If None, the error will be printed in the terminal.
            Default is None.
        """

        _stdout = None
        _stderr = None

        if out_file:
            _stdout = open(out_file, "w")
        if err_file:
            _stderr = open(err_file, "w")

        s = subprocess.Popen(str_cmd, shell=True, stdout=_stdout, stderr=_stderr)
        s.wait()

        if out_file:
            _stdout.flush()
            _stdout.close()
        if err_file:
            _stderr.flush()
            _stderr.close()

    def list_available_launchers(self) -> dict:
        """
        List the available launchers.

        Returns
        -------
        dict
            A list with the available launchers.
        """

        if hasattr(self, "available_launchers"):
            return self.available_launchers
        else:
            raise AttributeError("The attribute available_launchers is not defined.")

    def set_cases_dirs_from_output_dir(self) -> None:
        """
        Set the cases directories from the output directory.
        """

        if self.cases_dirs:
            self.logger.warning("Cases directories already set... resetting.")

        self.cases_dirs = sorted(
            [
                os.path.join(self.output_dir, case_dir)
                for case_dir in os.listdir(self.output_dir)
            ]
        )

        self.logger.info(f"Cases directories set from {self.output_dir}.")

    def write_array_in_file(self, array: np.ndarray, filename: str) -> None:
        """
        Write an array in a file.

        Parameters
        ----------
        array : np.ndarray
            The array to be written. Can be 1D or 2D.
        filename : str
            The name of the file.
        """

        write_array_in_file(array=array, filename=filename)

    def copy_files(self, src: str, dst: str) -> None:
        """
        Copy file(s) from source to destination.

        Parameters
        ----------
        src : str
            The source file.
        dst : str
            The destination file.
        """

        copy_files(src=src, dst=dst)

    def render_file_from_template(
        self, template_name: str, context: dict, output_filename: str = None
    ) -> None:
        """
        Render a file from a template.

        Parameters
        ----------
        template_name : str
            The name of the template file.
        context : dict
            The context to be used in the template.
        output_filename : str, optional
            The name of the output file. If None, it will be saved in the output
            directory with the same name as the template.
            Default is None.
        """

        template = self.env.get_template(name=template_name)
        rendered_content = template.render(context)
        if output_filename is None:
            output_filename = os.path.join(self.output_dir, template_name)
        with open(output_filename, "w") as f:
            f.write(rendered_content)

    def create_cases_context_one_by_one(self) -> List[dict]:
        """
        Create an array of dictionaries with the combinations of values from the
        input dictionary, one by one.

        Returns
        -------
        List[dict]
            A list of dictionaries, each representing a unique combination of
            parameter values.
        """

        num_cases = len(next(iter(self.model_parameters.values())))
        array_of_contexts = []
        for param, values in self.model_parameters.items():
            if len(values) != num_cases:
                raise ValueError(
                    f"All parameters must have the same number of values in one_by_one mode, check {param}"
                )

        for case_num in range(num_cases):
            case_context = {
                param: values[case_num]
                for param, values in self.model_parameters.items()
            }
            array_of_contexts.append(case_context)

        return array_of_contexts

    def create_cases_context_all_combinations(self) -> List[dict]:
        """
        Create an array of dictionaries with each possible combination of values
        from the input dictionary.

        Returns
        -------
        List[dict]
            A list of dictionaries, each representing a unique combination of
            parameter values.
        """

        keys = self.model_parameters.keys()
        values = self.model_parameters.values()
        combinations = itertools.product(*values)

        array_of_contexts = [
            dict(zip(keys, combination)) for combination in combinations
        ]

        return array_of_contexts

    def build_cases(self, mode: str = "one_by_one") -> None:
        """
        Create the cases folders and render the input files.

        Parameters
        ----------
        mode : str, optional
            The mode to create the cases. Can be "all_combinations" or "one_by_one".
            Default is "one_by_one".
        """

        if mode == "all_combinations":
            self.cases_context = self.create_cases_context_all_combinations()
        elif mode == "one_by_one":
            self.cases_context = self.create_cases_context_one_by_one()
        else:
            raise ValueError(f"Invalid mode to create cases: {mode}")
        for case_num, case_context in enumerate(self.cases_context):
            case_context["case_num"] = case_num
            case_dir = os.path.join(self.output_dir, f"{case_num:04}")
            self.cases_dirs.append(case_dir)
            os.makedirs(case_dir, exist_ok=True)
            for template_name in self.templates_name:
                self.render_file_from_template(
                    template_name=template_name,
                    context=case_context,
                    output_filename=os.path.join(case_dir, template_name),
                )
        self.logger.info(
            f"{len(self.cases_dirs)} cases created in {mode} mode and saved in {self.output_dir}"
        )

    def run_case(
        self,
        case_dir: str,
        launcher: str,
        ouput_log_file: str = "wrapper_out.log",
        error_log_file: str = "wrapper_error.log",
    ) -> None:
        """
        Run the case based on the launcher specified.

        Parameters
        ----------
        case_dir : str
            The case directory.
        launcher : str
            The launcher to run the case.
        ouput_log_file : str, optional
            The name of the output log file. Default is "wrapper_out.log".
        error_log_file : str, optional
            The name of the error log file. Default is "wrapper_error.log".
        """

        # Get launcher command from the available launchers
        launcher = self.list_available_launchers().get(launcher, launcher)

        # Run the case in the case directory
        self.logger.info(f"Running case in {case_dir} with launcher={launcher}.")
        os.chdir(case_dir)
        self._exec_bash_commands(
            str_cmd=launcher,
            out_file=ouput_log_file,
            err_file=error_log_file,
        )

    def run_cases(
        self,
        launcher: str,
        parallel: bool = False,
        cases_to_run: List[int] = None,
    ) -> None:
        """
        Run the cases based on the launcher specified.
        Parallel execution is optional.
        Cases to run can be specified.

        Parameters
        ----------
        launcher : str
            The launcher to run the cases.
        parallel : bool, optional
            If True, the cases will be run in parallel. Default is False.
        cases_to_run : List[int], optional
            The list with the cases to run. Default is None.
        """

        # Get launcher command from the available launchers
        launcher = self.list_available_launchers().get(launcher, launcher)

        if cases_to_run is not None:
            self.logger.warning(
                f"Cases to run was specified, so just {cases_to_run} will be run."
            )
            cases_dir_to_run = [self.cases_dirs[case] for case in cases_to_run]
        else:
            cases_dir_to_run = copy.deepcopy(self.cases_dirs)

        if parallel:
            num_threads = self.get_num_processors_available()
            self.logger.debug(
                f"Running cases in parallel with launcher={launcher}. Number of threads: {num_threads}."
            )
            with ThreadPoolExecutor(max_workers=num_threads) as executor:
                future_to_case = {
                    executor.submit(self.run_case, case_dir, launcher): case_dir
                    for case_dir in cases_dir_to_run
                }
                for future in as_completed(future_to_case):
                    case_dir = future_to_case[future]
                    try:
                        future.result()
                    except Exception as exc:
                        self.logger.error(
                            f"Job for {case_dir} generated an exception: {exc}."
                        )
        else:
            self.logger.info(f"Running cases sequentially with launcher={launcher}.")
            for case_dir in cases_dir_to_run:
                try:
                    self.run_case(
                        case_dir=case_dir,
                        launcher=launcher,
                    )
                except Exception as exc:
                    self.logger.error(
                        f"Job for {case_dir} generated an exception: {exc}."
                    )

        if launcher == "docker" or "docker" in launcher:
            # TODO: ALWAYS remove ALL stopped containers after running all cases
            remove_stopped_containers_cmd = 'docker ps -a --filter "ancestor=tausiaj/swash-image:latest" -q | xargs docker rm'
            self._exec_bash_commands(str_cmd=remove_stopped_containers_cmd)

        self.logger.info("All cases ran successfully.")

    def run_cases_bulk(
        self,
        launcher: str,
    ) -> None:
        """
        Run the cases based on the launcher specified.

        Parameters
        ----------
        launcher : str
            The launcher to run the cases.
        """

        self.logger.info(f"Running cases with launcher={launcher}.")
        os.chdir(self.output_dir)
        self._exec_bash_commands(str_cmd=launcher)

    def postprocess_case(self, case_num: int, case_dir: str) -> None:
        """
        Postprocess the model output.

        Parameters
        ----------
        case_num : int
            The case number.
        case_dir : str
            The case directory.
        """

        raise NotImplementedError("The method postprocess_case must be implemented.")

    def join_postprocessed_files(
        self, postprocessed_files: List[xr.Dataset]
    ) -> xr.Dataset:
        """
        Join the postprocessed files.

        Parameters
        ----------
        postprocessed_files : List[xr.Dataset]
            The list with the postprocessed files.
        """

        raise NotImplementedError(
            "The method join_postprocessed_files must be implemented."
        )

    def postprocess_cases(
        self, cases_to_postprocess: List[int] = None
    ) -> Union[xr.Dataset, List[xr.Dataset]]:
        """
        Postprocess the model output.

        Parameters
        ----------
        cases_to_postprocess : List[int], optional
            The list with the cases to postprocess. Default is None.

        Returns
        -------
        xr.Dataset or List[xr.Dataset]
            The postprocessed file or the list with the postprocessed files.
        """

        # TODO: Check if this option is necessary
        if os.path.exists(os.path.join(self.output_dir, "output_postprocessed.nc")):
            self.logger.warning(
                "Output postprocessed file already exists. Skipping postprocessing."
            )
            return xr.open_dataset(
                os.path.join(self.output_dir, "output_postprocessed.nc")
            )

        if not self.cases_dirs:
            self.logger.warning(
                "Cases directories are not set and will be searched from the output directory."
            )
            self.set_cases_dirs_from_output_dir()

        if cases_to_postprocess is not None:
            self.logger.warning(
                f"Cases to postprocess was specified, so just {cases_to_postprocess} will be postprocessed."
            )
            cases_dir_to_postprocess = [
                self.cases_dirs[case] for case in cases_to_postprocess
            ]
        else:
            cases_to_postprocess = list(range(len(self.cases_dirs)))
            cases_dir_to_postprocess = copy.deepcopy(self.cases_dirs)

        postprocessed_files = []
        for case_num, case_dir in zip(cases_to_postprocess, cases_dir_to_postprocess):
            self.logger.info(f"Postprocessing case {case_num} in {case_dir}.")
            try:
                postprocessed_file = self.postprocess_case(
                    case_num=case_num, case_dir=case_dir
                )
                postprocessed_files.append(postprocessed_file)
            except Exception as e:
                self.logger.error(
                    f"Ouput not postprocessed for case {case_num}. Error: {e}."
                )

        try:
            return self.join_postprocessed_files(
                postprocessed_files=postprocessed_files
            )
        except NotImplementedError as exc:
            self.logger.error(f"Error joining postprocessed files: {exc}")
            return postprocessed_files



import os.path as op

import wavespectra
import xarray as xr
from wavespectra.construct import construct_partition

from bluemath_tk.topo_bathy.swan_grid import generate_grid_parameters
from bluemath_tk.waves.binwaves import (
    process_kp_coefficients,
    transform_CAWCR_WS,
)
from bluemath_tk.wrappers.swan.swan_wrapper import SwanModelWrapper

example_directions = [
    7.5,
    22.5,
    37.5,
    52.5,
    67.5,
    82.5,
    97.5,
    112.5,
    127.5,
    142.5,
    157.5,
    172.5,
    187.5,
    202.5,
    217.5,
    232.5,
    247.5,
    262.5,
    277.5,
    292.5,
    307.5,
    322.5,
    337.5,
    352.5,
]
example_frequencies = [
    0.035,
    0.0385,
    0.042349998,
    0.046585,
    0.051243503,
    0.05636785,
    0.062004633,
    0.068205096,
    0.07502561,
    0.082528174,
    0.090780996,
    0.099859096,
    0.10984501,
    0.120829515,
    0.13291247,
    0.14620373,
    0.1608241,
    0.17690653,
    0.19459718,
    0.21405691,
    0.2354626,
    0.25900885,
    0.28490975,
    0.31340075,
    0.3447408,
    0.37921488,
    0.4171364,
    0.45885003,
    0.50473505,
]


class BinWavesWrapper(SwanModelWrapper):
    """ """

    def build_case(self, case_dir: str, case_context: dict):
        self.logger.info(f"Saving spectrum for {case_dir}")
        input_spectrum = construct_partition(
            freq_name="jonswap",
            freq_kwargs={
                "freq": example_frequencies,
                "fp": 1.0 / case_context.get("tp"),
                "hs": case_context.get("hs"),
            },
            dir_name="cartwright",
            dir_kwargs={
                "dir": example_directions,
                "dm": case_context.get("dir"),
                "dspr": case_context.get("spr"),
            },
        )
        wavespectra.SpecDataset(input_spectrum.to_dataset(name="efth")).to_swan(
            op.join(case_dir, "input_spectra.bnd")
        )

    def build_cases(self, mode="one_by_one"):
        super().build_cases(mode)
        for case_dir, case_context in zip(self.cases_dirs, self.cases_context):
            self.build_case(case_dir, case_context)


# Usage example
if __name__ == "__main__":
    # Define the input templates and output directory
    templates_dir = (
        "/home/tausiaj/GitHub-GeoOcean/BlueMath/bluemath_tk/wrappers/swan/templates/"
    )
    templates_name = ["input.swn", "depth_main.dat", "buoys.loc"]
    output_dir = "/home/tausiaj/GitHub-GeoOcean/BlueMath/test_cases/swan/javi/"
    # Load swan model parameters
    model_parameters = (
        xr.open_dataset("/home/tausiaj/GitHub-GeoOcean/BlueMath/test_data/subset.nc")
        .to_dataframe()
        .iloc[::60]
        .to_dict(orient="list")
    )
    # Create an instance of the SWAN model wrapper
    swan_wrapper = BinWavesWrapper(
        templates_dir=templates_dir,
        templates_name=templates_name,
        model_parameters=model_parameters,
        output_dir=output_dir,
    )
    # Build the input files
    swan_wrapper.build_cases(mode="one_by_one")
    # List available launchers
    print(swan_wrapper.list_available_launchers())
    # Run the model
    swan_wrapper.run_cases(launcher="docker", parallel=True)
    # Post-process the output files
    postprocessed_ds = swan_wrapper.postprocess_cases()
    postprocessed_ds.to_netcdf(op.join(swan_wrapper.output_dir, "waves_part.nc"))
    print(postprocessed_ds)
    # # Load spectra example
    # spectra = xr.open_dataset(
    #     "/home/tausiaj/GitHub-GeoOcean/BlueMath/test_data/Waves_Cantabria_356.08_43.82.nc"
    # )
    # spectra_transformed = transform_CAWCR_WS(spectra)
    # # Extract binwaves kp coeffs
    # kp_coeffs = process_kp_coefficients(
    #     swan_ds=postprocessed_ds,
    #     spectrum_freq=spectra_transformed.freq.values,
    #     spectrum_dir=spectra_transformed.dir.values,
    #     latitude=43.3,
    #     longitude=173.0,
    # )
    # print(kp_coeffs)



from typing import Tuple

import numpy as np
import pandas as pd
import xarray as xr
from matplotlib import cm
from matplotlib import pyplot as plt
from matplotlib.colors import ListedColormap

from ..core.plotting.base_plotting import DefaultStaticPlotting


def transform_CAWCR_WS(cawcr_dataset: xr.Dataset) -> xr.Dataset:
    """
    Transform the wave spectra from CAWCR format to binwaves format.

    Parameters
    ----------
    cawcr_dataset : xr.Dataset
        The wave spectra dataset in CAWCR format.

    Returns
    -------
    xr.Dataset
        The wave spectra dataset in binwaves format.
    """

    ds = cawcr_dataset.rename({"frequency": "freq", "direction": "dir"})
    ds["efth"] = ds["efth"] * np.pi / 180.0
    ds["dir"] = ds["dir"] - 180.0
    ds["dir"] = np.where(ds["dir"] < 0, ds["dir"] + 360, ds["dir"])

    return ds


def process_kp_coefficients(
    swan_ds: xr.Dataset,
    spectrum_freq: np.ndarray,
    spectrum_dir: np.ndarray,
    latitude: float,
    longitude: float,
):
    """
    This function processes the propagation coefficients for all the grid points
    within the SWAN simulation output.
    It takes a long time to run but it only needs to be done once per location.

    Parameters
    ----------
    swan_ds : xr.Dataset
        The SWAN processed dataset.
    spectrum_freq : np.ndarray
        The frequency array.
    spectrum_dir : np.ndarray
        The direction array.
    latitude : float
        The latitude of the point of interest.
    longitude : float
        The longitude of the point of interest.

    Returns
    -------
    xr.Dataset
        The propagation coefficients dataset.
    """

    kp_matrix = np.full(
        [
            len(swan_ds.case_num),
            len(spectrum_freq),
            len(spectrum_dir),
        ],
        0.0,
    )

    swan_point = swan_ds.sel(
        Xp=longitude,
        Yp=latitude,
        method="nearest",
    )
    # TODO: Check if this is the correct way to handle NaN values
    if any(np.isnan(swan_point["TPsmoo"].values)):
        raise ValueError("NaN values found for variable TPsmoo_part")

    # Tp mask
    swan_point_cut = swan_point[["Hsig", "TPsmoo", "Dir"]].where(
        swan_point["TPsmoo"] > 0,
        drop=True,
    )

    # get k,f,d
    kfd = xr.Dataset(
        {
            "k": swan_point_cut.Hsig,
            "f": 1 / swan_point_cut.TPsmoo,
            "d": swan_point_cut.Dir,
        }
    )

    # fill kp
    for case_num in kfd.case_num.values:
        kfd_c = kfd.sel(case_num=case_num)

        # get k,f,d and clean nans
        k = kfd_c.k.values
        f = kfd_c.f.values
        d = kfd_c.d.values

        k = k[~np.isnan(f)]
        d = d[~np.isnan(f)]
        f = f[~np.isnan(f)]

        # set case kp at point
        for c in range(len(f)):
            i = np.argmin(np.abs(spectrum_freq - f[c]))
            j = np.argmin(np.abs(spectrum_dir - d[c]))
            kp_matrix[case_num, i, j] = k[c]

    return xr.Dataset(
        {
            "kp": (["case", "freq", "dir"], kp_matrix),
            "swan_freqs": (["case"], 1.0 / swan_point_cut.TPsmoo),
            "swan_dirs": (["case"], swan_point_cut.Dir),
        },
        coords={
            "case": swan_ds.case_num,
            "freq": spectrum_freq,
            "dir": spectrum_dir,
        },
    )


def reconstruc_spectra(
    spectra_ds: xr.Dataset,
    kp_coeffs: xr.Dataset,
):
    """
    Reconstruct the wave spectra using the kp coefficients.

    Parameters
    ----------
    spectra_ds : xr.Dataset
        The offshore wave spectra dataset.
    kp_coeffs : xr.Dataset
        The nearshore kp coefficients dataset.

    Returns
    -------
    xr.Dataset
        The nearshore reconstructed wave spectra dataset.
    """

    EFTH = np.full(
        np.shape(spectra_ds.efth.values),
        0,
    )

    for case in range(len(kp_coeffs.case)):
        freq_, dir_ = (
            kp_coeffs.isel(case=case).swan_freqs.values,
            kp_coeffs.isel(case=case).swan_dirs.values,
        )
        efth_case = spectra_ds.sel(freq=freq_, dir=dir_, method="nearest")
        kp_case = kp_coeffs.sortby("dir").isel(case=case)

        EFTH = EFTH + (efth_case.efth * kp_case.kp**2).values

    # ns_sp = off_sp.drop(("Wspeed", "Wdir", "Depth")).copy()

    return xr.Dataset(
        {
            "efth": (["time", "freq", "dir"], EFTH),
        },
        coords={
            "time": spectra_ds.time,
            "freq": spectra_ds.freq,
            "dir": spectra_ds.dir,
        },
    )


def plot_selected_subset_parameters(
    selected_subset: pd.DataFrame,
    color: str = "blue",
    **kwargs,
) -> Tuple[plt.figure, plt.axes]:
    """
    Plot the selected subset parameters.

    Parameters
    ----------
    selected_subset : pd.DataFrame
        The selected subset parameters.
    color : str, optional
        The color to use in the plot. Default is "blue".
    **kwargs : dict, optional
        Additional keyword arguments to be passed to the scatter plot function.

    Returns
    -------
    plt.figure
        The figure object containing the plot.
    plt.axes
        Array of axes objects for the subplots.
    """

    # Create figure and axes
    default_static_plot = DefaultStaticPlotting()
    fig, axes = default_static_plot.get_subplots(
        nrows=len(selected_subset) - 1,
        ncols=len(selected_subset) - 1,
        sharex=False,
        sharey=False,
    )

    for c1, v1 in enumerate(list(selected_subset.columns)[1:]):
        for c2, v2 in enumerate(list(selected_subset.columns)[:-1]):
            default_static_plot.plot_scatter(
                ax=axes[c2, c1],
                x=selected_subset[v1],
                y=selected_subset[v2],
                c=color,
                alpha=0.6,
                **kwargs,
            )
            if c1 == c2:
                axes[c2, c1].set_xlabel(list(selected_subset.columns)[c1 + 1])
                axes[c2, c1].set_ylabel(list(selected_subset.columns)[c2])
            elif c1 > c2:
                axes[c2, c1].xaxis.set_ticklabels([])
                axes[c2, c1].yaxis.set_ticklabels([])
            else:
                fig.delaxes(axes[c2, c1])

    return fig, axes


def plot_grid_cases(spectra: xr.Dataset, cases_id: np.ndarray, figsize: tuple = (8, 8)):
    """
    Function to plot the cases with different colors.

    Parameters
    ----------
    spectra : xr.Dataset
        The wave spectra dataset.
    cases_id : np.ndarray
        The cases ids for the color.
    figsize : tuple, optional
        The figure size. Default is (8, 8).

    Returns
    -------
    plt.Figure
        The figure.
    """

    # generate figure and axes
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(1, 1, 1, projection="polar")

    # prepare data
    x = np.append(
        np.deg2rad(spectra.dir.values - 7.5), np.deg2rad(spectra.dir.values - 7.5)[0]
    )
    y = np.append(0, spectra.freq.values)
    z = cases_id

    # custom colormap
    cmn = np.vstack(
        (
            cm.get_cmap("plasma", 124)(np.linspace(0, 0.9, 70)),
            cm.get_cmap("magma_r", 124)(np.linspace(0.1, 0.4, 80)),
            cm.get_cmap("rainbow_r", 124)(np.linspace(0.1, 0.8, 80)),
            cm.get_cmap("Blues_r", 124)(np.linspace(0.4, 0.8, 40)),
            cm.get_cmap("cubehelix_r", 124)(np.linspace(0.1, 0.8, 80)),
        )
    )
    cmn = ListedColormap(cmn, name="cmn")

    # plot cases id
    p1 = ax.pcolormesh(
        x,
        y,
        z,
        vmin=0,
        vmax=np.nanmax(cases_id),
        edgecolor="grey",
        linewidth=0.005,
        cmap=cmn,
        shading="flat",
    )

    # customize axes
    ax.set_theta_zero_location("N", offset=0)
    ax.set_theta_direction(-1)
    ax.tick_params(
        axis="both",
        colors="black",
        labelsize=14,
        pad=10,
    )

    # add colorbar
    plt.colorbar(p1, pad=0.1, shrink=0.7).set_label("Case ID", fontsize=16)

    return fig
